"""Covers import of data downloaded from the
`Meadows online behavior platform <https://meadows-research.com/>`_.


For information on available file types see the meadows
`documentation on downloads <https://meadows-research.com/documentation\
/researcher/downloads/>`_.
"""
from os.path import basename
import numpy
from scipy.io import loadmat
from pyrsa.rdm.rdms import RDMs


def load_rdms(fpath, sort=True):
    """Read a Meadows results file and return any RDMs as a pyrsa object

    Args:
        fpath (str): path to .mat Meadows results file
        sort (bool): whether to sort the RDM based on the stimulus names

    Raises:
        ValueError: Will raise an error if the file is missing an expected
            variable. This can happen if the file does not contain MA task
            data.

    Returns:
        RDMs: All rdms found in the data file as an RDMs object
    """
    info = extract_filename_segments(fpath)
    data = loadmat(fpath)
    if info['participant_scope'] == 'single':
        for var in ('stimuli', 'rdmutv'):
            if var not in data:
                raise ValueError(f'File missing variable: {var}')
        utvs = data['rdmutv']
        stimuli_fnames = data['stimuli']
        pnames = [info['participant']]
    else:
        stim_vars = [v for v in data.keys() if v[:7] == 'stimuli']
        stimuli_fnames = data[stim_vars[0]]
        pnames = ['-'.join(v.split('_')[1:]) for v in stim_vars]
        utv_vars = ['rdmutv_' + p.replace('-', '_') for p in pnames]
        utvs = numpy.squeeze(numpy.stack([data[v] for v in utv_vars]))

    desc_info_keys = (
        'participant',
        'task_index',
        'task_name',
        'experiment_name'
    )
    conds = [f.split('.')[0] for f in stimuli_fnames]
    rdms = RDMs(
        utvs,
        dissimilarity_measure='euclidean',
        descriptors={k: info[k] for k in desc_info_keys if k in info},
        rdm_descriptors=dict(participants=pnames),
        pattern_descriptors=dict(conds=conds),
    )
    if sort:
        rdms.sort_by(conds='alpha')
    return rdms


def extract_filename_segments(fpath):
    """Get information from the name of a downloaded results file

    Will determine:
        * participant_scope: 'single' or 'multiple', how many participant
            sessions this file covers.
        * task_scope: 'single' or 'multiple', how many experiment tasks this
            file covers.
        * participant: the Meadows nickname of the participant, if this is a
            single participation file.
        * task_index: the 1-based index of the task in the experiment, if
            this is a single participant file.
        * task_name: the name of the task in the experiment, if
            this is not a single participant file.
        * version: the experiment version as a string.
        * experiment_name: name of the experiment on Meadows.
        * structure: the structure of the data contained, one of 'tree',
            'events', '1D', '2D', etc.
        * filetype: the file extension and file format used to serialize the
            data.

    Args:
        fpath (str): File system path to downloaded file

    Returns:
        dict: Dictionary with the fields described above.
    """
    fname, ext = basename(fpath).split('.')
    segments = fname.split('_')
    info = dict(
        task_scope='single',
        version=segments[3].replace('v', ''),
        experiment_name=segments[1],
        structure=segments[-1],
        filetype=ext
    )
    if segments[-2].isdigit():
        info['participant_scope'] = 'single'
        info['participant'] = segments[-3]
        info['task_index'] = int(segments[-2])
    else:
        info['participant_scope'] = 'multiple'
        info['task_name'] = segments[-2]
    return info



"""Unit tests for expanding RDMs to include missing data.
"""
#pylint: disable=import-outside-toplevel
from unittest import TestCase
from numpy import array, nan
from numpy.testing import assert_array_equal


class RdmsExpandTests(TestCase):
    """Unit tests for expanding RDMs to include missing data.
    """

    def test_expand_based_on_list_of_rdms_objects(self):
        """In this case the complete list of conditions is determined
        from the RDMs passed.
        """
        from meadows.brdms import BehavioralRDMs as RDMs
        rdms1 = RDMs(
            dissimilarities=array([[1, 2, 3]]),
            dissimilarity_measure='shared_measure',
            descriptors=dict(shared_desc='shared_val', diff_desc='one'),
            rdm_descriptors=dict(rdesc=['foo1']),
            pattern_descriptors=dict(conds=['a', 'b', 'c']),
        )
        rdms23 = RDMs(
            dissimilarities=array([[4, 5, 6], [7, 8, 9]]),
            dissimilarity_measure='shared_measure',
            descriptors=dict(shared_desc='shared_val', diff_desc='two-three'),
            rdm_descriptors=dict(rdesc=['foo2', 'foo3']),
            pattern_descriptors=dict(conds=['b', 'c', 'd']),
        )
        rdms = RDMs.expand([rdms1, rdms23])
        self.assertEqual(rdms.n_rdm, 3)
        self.assertEqual(rdms.n_cond, 4)
        self.assertEqual(rdms.dissimilarity_measure, 'shared_measure')
        self.assertEqual(rdms.descriptors.get('shared_desc'), 'shared_val')
        assert_array_equal(
            rdms.rdm_descriptors.get('diff_desc'),
            ['one', 'two-three', 'two-three']
        )
        assert_array_equal(
            rdms.rdm_descriptors.get('rdesc'),
            ['foo1', 'foo2', 'foo3']
        )
        assert_array_equal(
            rdms.pattern_descriptors.get('conds'),
            ['a', 'b', 'c', 'd']
        )
        assert_array_equal(
            rdms.dissimilarities,
            array([
                [  1,   2, nan,   3, nan, nan],
                [nan, nan, nan,   4,   5,   6],
                [nan, nan, nan,   7,   8,   9]
            ])
        )

    def test_expand_with_list_of_pattern_descriptors(self):
        """Where the user explicitly chooses the patterns

        We pass a list with a single RDMs object containing one RDM,
        then specify one additional pattern not covered in the RDM.
        """
        from meadows.brdms import BehavioralRDMs as RDMs
        rdms1 = RDMs(
            dissimilarities=array([[1, 2, 3]]),
            dissimilarity_measure='measure',
            pattern_descriptors=dict(conds=['b', 'c', 'd']),
        )
        rdms = RDMs.expand([rdms1], all_patterns=['a', 'b', 'c', 'd'])
        self.assertEqual(rdms.n_rdm, 1)
        self.assertEqual(rdms.n_cond, 4)
        assert_array_equal(
            rdms.pattern_descriptors.get('conds'),
            ['a', 'b', 'c', 'd']
        )
        assert_array_equal(
            rdms.dissimilarities,
            array([
                [nan, nan, nan,   1,   2,   3],
            ])
        )



"""Unit tests for aligning and averaging partial RDMs
"""
#pylint: disable=import-outside-toplevel, no-self-use
from unittest import TestCase
from numpy import array, nan, isnan
from numpy.testing import assert_almost_equal, assert_array_equal
from scipy.stats import pearsonr


def non_nan(vectors, row=0):
    """Select only non-nan values from the first row"""
    return vectors[row, ~isnan(vectors[row, :])]


class RdmsAlignTests(TestCase):
    """Unit tests for aligning and averaging partial RDMs
    """

    def test_align(self):
        """The align method bring the RDMs as close together as possible
        """
        from meadows.brdms import BehavioralRDMs as RDMs
        partial=array([
            [  1,   2, nan,   3, nan, nan],
            [nan, nan, nan,   4,   5,   6],
        ])
        partial_rdms = RDMs(
            dissimilarities=partial
        )
        aligned_rdms = partial_rdms.align()
        aligned = aligned_rdms.dissimilarities
        assert_almost_equal(aligned[0, 3], aligned[1, 3], decimal=4)
        assert_almost_equal(
            pearsonr(non_nan(partial), non_nan(aligned))[0],
            1,
            decimal=7
        )
        actual_aligned = array([
            [0.1438, 0.2877, nan, 0.4315,    nan,    nan],
            [   nan,    nan, nan, 0.4316, 0.5395, 0.6474]
        ])
        assert_almost_equal(aligned, actual_aligned, decimal=4)
        assert_array_equal(
            aligned_rdms.rdm_descriptors.get('weights'),
            array([
                [  1,   4, nan,    9,  nan,  nan],
                [nan, nan, nan,   16,   25,   36],
            ])
        )

    def test_align_setsize(self):
        """The align method bring the RDMs as close together as possible
        """
        from meadows.brdms import BehavioralRDMs as RDMs
        partial_rdms = RDMs(
            dissimilarities=array([
                [  1,   2, nan,   3, nan, nan],
                [nan, nan, nan,   4,   5, nan],
            ])
        )
        aligned_rdms = partial_rdms.align(method='setsize')
        assert_almost_equal(
            aligned_rdms.rdm_descriptors.get('weights'),
            array([
                [0.3333, 0.3333, nan, 0.3333,  nan,  nan],
                [   nan,    nan, nan,    0.5,   0.5, nan],
            ]),
            decimal=4
        )

    def test_mean_no_weights(self):
        """RDMs.mean() returns an RDMs with the nan omitted mean of the rdms
        """
        from meadows.brdms import BehavioralRDMs as RDMs
        partial_rdms = RDMs(
            dissimilarities=array([
                [  1,   2, nan,   3, nan, nan],
                [  2,   1, nan,   4,   5,   6],
            ])
        )
        assert_almost_equal(
            partial_rdms.mean(weights=None).dissimilarities,
            array([[ 1.5,  1.5, nan, 3.5, 5, 6]])
        )

    def test_weighted_mean(self):
        """Weights passed or stored in a descriptor are used in average
        """
        from meadows.brdms import BehavioralRDMs as RDMs
        partial_rdms = RDMs(
            dissimilarities=array([
                [  1,   2, nan,   3, nan, nan],
                [  2,   1, nan,   4,   5,   6],
            ])
        )
        weights = array([
            [  1,   1, nan,   1, nan, nan],
            [  2,   2, nan,   2,   2,   2],
        ])
        assert_almost_equal(
            partial_rdms.mean(weights=weights).dissimilarities,
            array([[1.6667, 1.3333, nan, 3.6667, 5.0000, 6.0000]]),
            decimal=3
        )
        partial_rdms.rdm_descriptors['weights'] = weights
        assert_almost_equal(
            partial_rdms.mean(weights='stored').dissimilarities,
            array([[1.6667, 1.3333, nan, 3.6667, 5.0000, 6.0000]]),
            decimal=3
        )



from collections import Counter
from copy import deepcopy
import numpy as np
from scipy.spatial.distance import squareform
from meadows.align import _mean, _align
from pyrsa.rdm.rdms import RDMs


class BehavioralRDMs(RDMs):
    """Behavioral RDMs class

    Args:
        dissimilarities (numpy.ndarray):
            either a 2d np-array (n_rdm x vectorform of dissimilarities)
            or a 3d np-array (n_rdm x n_cond x n_cond)
        dissimilarity_measure (String):
            a description of the dissimilarity measure (e.g. 'Euclidean')
        descriptors (dict):
            descriptors with 1 value per RDMs object
        rdm_descriptors (dict):
            descriptors with 1 value per RDM
        pattern_descriptors (dict):
            descriptors with 1 value per RDM column

    Attributes:
        n_rdm(int): number of rdms
        n_cond(int): number of patterns
    """

    def reorder(self, new_order):
        """Reorder the patterns according to the index in new_order

        Args:
            new_order (numpy.ndarray): new order of patterns,
                vector of length equal to the number of patterns
        """
        matrices = self.get_matrices()
        matrices = matrices[(slice(None),) + np.ix_(new_order, new_order)]
        self.dissimilarities = batch_to_vectors(matrices)[0]
        for dname, descriptors in self.pattern_descriptors.items():
            self.pattern_descriptors[dname] = descriptors[new_order]

    def sort_by(self, **kwargs):
        """Reorder the patterns by sorting a descriptor

        Pass keyword arguments that correspond to descriptors,
        with value 'alpha'. 

        Example:
            Sorts the condition descriptor alphabetically:

            `rdms.sort(condition='alpha')`

        Raises:
            ValueError: Raised if the method chosen is not implemented
        """
        for dname, method in kwargs.items():
            if method == 'alpha':
                descriptor = self.pattern_descriptors[dname]
                self.reorder(np.argsort(descriptor))
            else:
                raise ValueError(f'Unknown sorting method: {method}')

    @classmethod
    def expand(cls, list_of_rdms, all_patterns=None, descriptor='conds'):
        """Make larger RDMs with missing values where needed

        Any object-level descriptors will be turned into rdm_descriptors
        if they do not match across objects.

        Args:
            list_of_rdms (list): List of RDMs objects
            all_patterns (list, optional): The full list of pattern
                descriptors. Defaults to None, in which case the full
                list is the union of all input rdms' values for the
                pattern descriptor chosen.
            descriptor (str, optional): The pattern descriptor on the basis
                of which to expand. Defaults to 'conds'.

        Returns:
            RDMs: Object containing all input rdms on the larger scale,
                with missing values where required
        """

        def pdescs(rdms, descriptor):
            return list(rdms.pattern_descriptors.get(descriptor, []))

        if all_patterns is None:
            all_patterns = []
            for rdms in list_of_rdms:
                all_patterns += pdescs(rdms, descriptor)
            all_patterns = list(dict.fromkeys(all_patterns).keys())

        n_rdm_objs = len(list_of_rdms)
        n_rdms = sum([rdms.n_rdm for rdms in list_of_rdms])
        n_patterns = len(all_patterns)

        desc_tuples = []
        rdm_desc_names = []
        for rdms in list_of_rdms:
            desc_tuples += list(rdms.descriptors.items())
            rdm_desc_names += list(rdms.rdm_descriptors.keys())
        descriptors = dict(
            [d for d, c in Counter(desc_tuples).items() if c == n_rdm_objs]
        )
        desc_diff_names = set(
            [d[0] for d, c in Counter(desc_tuples).items() if c != n_rdm_objs]
        )
        rdm_desc_names = set(rdm_desc_names + list(desc_diff_names))
        rdm_descriptors = dict([(n, [None]*n_rdms) for n in rdm_desc_names])

        measure = None
        vector_len = int(n_patterns * (n_patterns-1) / 2)
        vectors = np.full((n_rdms, vector_len), np.nan)
        rdm_id = 0
        for rdms in list_of_rdms:
            measure = rdms.dissimilarity_measure
            pidx = [all_patterns.index(i) for i in pdescs(rdms, descriptor)]
            for rdm_local_id, utv in enumerate(rdms.dissimilarities):
                rdm = np.full((len(all_patterns), len(all_patterns)), np.nan)
                rdm[np.ix_(pidx, pidx)] = squareform(utv, checks=False)
                vectors[rdm_id, :] = squareform(rdm, checks=False)
                for name in rdm_descriptors.keys():
                    if name in rdms.rdm_descriptors:
                        val = rdms.rdm_descriptors[name][rdm_local_id]
                        rdm_descriptors[name][rdm_id] = val
                    elif name in rdms.descriptors:
                        rdm_descriptors[name][rdm_id] = rdms.descriptors[name]
                rdm_id += 1

        return RDMs(
            dissimilarities=vectors,
            dissimilarity_measure=measure,
            descriptors=descriptors,
            rdm_descriptors=rdm_descriptors,
            pattern_descriptors=dict([(descriptor, all_patterns)])
        )

    def align(self, method='evidence'):
        """Bring RDMs closer together

        Iteratively scales RDMs based on pairs in-common.
        Also adds an RDM descriptor with the weights used.

        Args:
            method (str, optional): One of 'evidence', 'setsize' or
                'simple'. Defaults to 'evidence'.

        Returns:
            RDMs: RDMs object with the aligned RDMs
        """
        aligned, weights = _align(self.dissimilarities, method)
        rdm_descriptors = deepcopy(self.rdm_descriptors)
        if weights is not None:
            rdm_descriptors['weights'] = weights
        return RDMs(
            dissimilarities=aligned,
            dissimilarity_measure=self.dissimilarity_measure,
            descriptors=deepcopy(self.descriptors),
            rdm_descriptors=rdm_descriptors,
            pattern_descriptors=deepcopy(self.pattern_descriptors)
        )

    def mean(self, weights='stored'):
        """Average rdm of all rdms contained

        Args:
            weights (str or ndarray, optional): One of:
                'stored': Use the weights contained in `rdm_descriptor` "weights"
                ndarray: Weights array of the shape of RDMs.dissimilarities
                None: No weighting applied

        Returns:
            `pyrsa.rdm.rdms.RDMs`: New RDMs object with one vector
        """
        if weights == 'stored':
            weights = self.rdm_descriptors.get('weights')
        new_descriptors = dict(
            [(k, v) for (k, v) in self.descriptors.items() if k != 'weights']
        )
        return RDMs(
            dissimilarities=np.array([_mean(self.dissimilarities, weights)]),
            dissimilarity_measure=self.dissimilarity_measure,
            descriptors=new_descriptors,
            pattern_descriptors=deepcopy(self.pattern_descriptors)
        )



"""Private implementations of align() and mean()
"""
import numpy as np
from numpy import sqrt, nan, inf


def _mean(vectors, weights=None):
    """Weighted mean of RDM vectors, ignores nans

    See :meth:`pyrsa.rdm.rdms.RDMs.mean`

    Args:
        vectors (ndarray): dissimilarity vectors of shape (nrdms, nconds)
        weights (ndarray, optional): Same shape as vectors.

    Returns:
        ndarray: Average vector of shape (nconds,)
    """
    if weights is None:
        weights = np.ones(vectors.shape)
        weights[np.isnan(vectors)] = np.nan
    weighted_sum = np.nansum(vectors * weights, axis=0)
    return weighted_sum / np.nansum(weights, axis=0)


def _ss(vectors):
    """Sum of squares on the last dimension

    Args:
        vectors (ndarray): 1- or 2-dimensional data

    Returns:
        ndarray: the sum of squares, with an extra empty dimension
    """
    summed_squares = np.nansum(vectors ** 2, axis=vectors.ndim-1)
    return np.expand_dims(summed_squares, axis=vectors.ndim-1)


def _scale(vectors):
    """Divide by the root sum of squares

    Args:
        vectors (ndarray): 1- or 2-dimensional data

    Returns:
        ndarray: input scaled
    """
    return vectors / sqrt(_ss(vectors))


def _align(dissim, method):
    """Align RDM vectors

    See :meth:`pyrsa.rdm.rdms.RDMs.align`

    Args:
        dissim (ndarray): dissimilarity vectors, shape = (rdms, conds)
        method (str): one of 'evidence', 'setsize' or 'simple'.

    Returns:
        (ndarray, ndarray): Tuple of the aligned dissimilarity vectors
            and the weights used
    """
    n_rdms, n_conds = dissim.shape
    if method == 'evidence':
        weights = (dissim ** 2).clip(0.2)
    elif method == 'setsize':
        setsize = np.isfinite(dissim).sum(axis=1)
        weights = np.tile(1 / setsize, [n_conds, 1]).T
    else:
        weights = np.ones(dissim.shape)
    weights[np.isnan(dissim)] = np.nan

    current_estimate = _scale(_mean(dissim))
    prev_estimate = np.full([n_conds,], -inf)
    while _ss(current_estimate - prev_estimate) > 1e-8:
        prev_estimate = current_estimate.copy()
        tiled_estimate = np.tile(current_estimate, [n_rdms, 1])
        tiled_estimate[np.isnan(dissim)] = nan
        aligned = _scale(dissim) * sqrt(_ss(tiled_estimate))
        current_estimate = _scale(_mean(aligned, weights))

    return aligned, weights



from unittest import TestCase
import pkg_resources
from numpy.testing import assert_array_equal, assert_array_almost_equal


class MeadowsIOTests(TestCase):
    """Acceptance and unit tests for loading Meadows data.
    """

    def test_load_rdms_from_mat_file_1p(self):
        """Acceptance test for loading data from a Meadows
        .mat file download containing data for a single task,
        single participant. Should have descriptors and dissimilarities
        as found in file.
        """
        import pyrsa.io.meadows
        fname = 'Meadows_myExp_v_v1_cuddly-bunny_3_1D.mat'
        fpath = pkg_resources.resource_filename('tests', 'data/' + fname)
        rdms = pyrsa.io.meadows.load_rdms(fpath, sort=False)
        self.assertEqual(rdms.descriptors.get('participant'), 'cuddly-bunny')
        self.assertEqual(rdms.descriptors.get('task_index'), 3)
        self.assertEqual(rdms.descriptors.get('experiment_name'), 'myExp')
        self.assertEqual(rdms.dissimilarity_measure, 'euclidean')
        conds = rdms.pattern_descriptors.get('conds')
        assert_array_equal(conds[:2], ['stim118', 'stim117'])
        assert_array_almost_equal(
            rdms.dissimilarities[0, :2],
            [0.00791285387561264, 0.00817090931233484]
        )

    def test_load_rdms_and_sort(self):
        """As above but with sorting
        """
        import pyrsa.io.meadows
        fname = 'Meadows_myExp_v_v1_cuddly-bunny_3_1D.mat'
        fpath = pkg_resources.resource_filename('tests', 'data/' + fname)
        rdms = pyrsa.io.meadows.load_rdms(fpath, sort=True)
        conds = rdms.pattern_descriptors.get('conds')
        assert_array_equal(conds[:2], ['stim001', 'stim002'])
        assert_array_equal(conds[-2:], ['stim117', 'stim118'])
        assert_array_almost_equal(
            rdms.dissimilarities[0, -2:],
            [0.00817090931233484, 0.00791285387561264]
        )

    def test_load_rdms_from_mat_file_3p(self):
        """Acceptance test for loading data from a Meadows
        .mat file download containing data for a single task,
        multiple participants. Should have descriptors and dissimilarities
        as found in file.
        """
        import pyrsa.io.meadows
        fname = 'Meadows_myExp_v_v1_arrangement_1D.mat'
        fpath = pkg_resources.resource_filename('tests', 'data/' + fname)
        rdms = pyrsa.io.meadows.load_rdms(fpath, sort=False)
        self.assertEqual(rdms.descriptors.get('task_name'), 'arrangement')
        self.assertEqual(rdms.descriptors.get('experiment_name'), 'myExp')
        self.assertEqual(
            rdms.rdm_descriptors.get('participants').tolist(),
            ['able-fly', 'clean-koi', 'cuddly-bunny']
        )
        self.assertEqual(rdms.dissimilarity_measure, 'euclidean')
        conds = rdms.pattern_descriptors.get('conds')
        assert_array_equal(conds[:2], ['stim118', 'stim117'])
        assert_array_almost_equal(
            rdms.dissimilarities[0, :2], ## 'able-fly'
            [0.0165981067918494, 0.0123233998529090]
        )
        assert_array_almost_equal(
            rdms.dissimilarities[1, :2], ## 'clean-koy'
            [0.00773234353884765, 0.00589909056106329]
        )

    def test_extract_filename_segments_1p_1t(self):
        """Test interpretation of the filename of a Meadows results download

        This case covers a filename for a single participant, single task,
        downloaded in 2019.
        """
        from pyrsa.io.meadows import extract_filename_segments
        fname = 'Meadows_myExp_v_v1_cuddly-bunny_3_1D.mat'
        info = extract_filename_segments(fname)
        self.assertEqual(info.get('participant_scope'), 'single')
        self.assertEqual(info.get('task_scope'), 'single')
        self.assertEqual(info.get('participant'), 'cuddly-bunny')
        self.assertEqual(info.get('task_index'), 3)
        self.assertEqual(info.get('version'), '1')
        self.assertEqual(info.get('experiment_name'), 'myExp')
        self.assertEqual(info.get('structure'), '1D')
        self.assertEqual(info.get('filetype'), 'mat')
        self.assertIsNone(info.get('task_name'))

    def test_extract_filename_segments_mp_1t(self):
        """Test interpretation of the filename of a Meadows results download

        This case covers a filename for multiple participants, single task,
        downloaded in 2019.
        """
        from pyrsa.io.meadows import extract_filename_segments
        fname = 'Meadows_myExp_v_v1_arrangement_1D.mat'
        info = extract_filename_segments(fname)
        self.assertEqual(info.get('participant_scope'), 'multiple')
        self.assertEqual(info.get('task_scope'), 'single')
        self.assertEqual(info.get('task_name'), 'arrangement')
        self.assertEqual(info.get('version'), '1')
        self.assertEqual(info.get('experiment_name'), 'myExp')
        self.assertEqual(info.get('structure'), '1D')
        self.assertEqual(info.get('filetype'), 'mat')
        self.assertIsNone(info.get('participant'))
        self.assertIsNone(info.get('task_index'))
